# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""The eqdsk package provides a class for reading and writing eqdsk files.

The core class is the [`EQDSKInterface`][eqdsk.file],
responsible for most of the functionality.

There is also the COCOS implementation in the [`cocos.py`][eqdsk.cocos] file,
which provides the coordinate system conversion functions.
"""

from eqdsk.file import EQDSKInterface



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
"""Eqdsk logging"""

import logging

logger = logging.getLogger("EQDSK Logger")

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)8s] %(message)s",
)


def eqdsk_warn(*args, **kwargs):
    """Warning"""
    stacklevel = kwargs.pop("stacklevel", 1) + 1
    logger.warning(*args, stacklevel=stacklevel, **kwargs)


def eqdsk_print(*args, **kwargs):
    """Printing"""
    stacklevel = kwargs.pop("stacklevel", 1) + 1
    logger.info(*args, stacklevel=stacklevel, **kwargs)


def eqdsk_banner():
    """
    Returns
    -------
    :
        The eqdsk banner
    """
    from eqdsk._version import version  # noqa: PLC0415

    return f"""
********************************
*    ___ ___  ___  ___ _  __   *
*   | __/ _ \\|   \\/ __| |/ /   *
*   | _| (_) | |) \\__ \\ ' <    *
*   |___\\__\\_\\___/|___/_|\\_\\   *
*                              *
*   The Bluemira Developers    *
********************************
EQDSK Version: {version}
"""



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later
"""Eqdsk CLI"""

from pathlib import Path

import click
import numpy as np

from eqdsk.cocos import COCOS, KnownCOCOS
from eqdsk.file import EQDSKInterface
from eqdsk.log import eqdsk_banner, eqdsk_warn


def _setup_plotting(eq: EQDSKInterface):
    try:
        import matplotlib.pyplot as plt  # noqa: PLC0415
    except ImportError:
        raise ImportError(
            "Matplotlib not installed, cannot plot.\n"
            "To resolve, try running:\n"
            "pip install matplotlib"
        ) from None

    fig, ax = plt.subplots()
    fig.set_size_inches(8, 8)

    fig.suptitle(f"EQDSK:  {eq.name}")
    ax.set_xlabel("R [m]")
    ax.set_ylabel("Z [m]")
    ax.set_aspect("equal")

    # plot mag center
    ax.plot(eq.xmag, eq.zmag, color="red", marker="o")

    return fig, ax, plt.show


@click.group()
@click.version_option()
def cli():
    """EQDSK cli

    Useful utilities for displaying eqdsk file data.
    """


@cli.command("show", no_args_is_help=True)
@click.argument("filepath", type=click.Path(exists=True))
def show(filepath):
    """Reads and prints important parameters of the eqdsk."""
    eq = EQDSKInterface.from_file(filepath, no_cocos=True)

    print(eqdsk_banner())  # noqa: T201
    print(eq)  # noqa: T201


@cli.command("plot", no_args_is_help=True)
@click.argument("filepath", type=click.Path(exists=True))
def plot_bdry(filepath):
    """
    Plot the eqdsk plasma boundary.

    matplotlib is required for plotting.
    """
    eq = EQDSKInterface.from_file(filepath, no_cocos=True)

    _fig, ax, show = _setup_plotting(eq)

    ax.plot(eq.xbdry, eq.zbdry, color="blue", linestyle="-", linewidth=2)
    ax.legend(["Magnetic center", "Plasma Boundary"], loc="upper right")

    zoom = 1.1
    plot_max_x = max(eq.xbdry) * zoom
    plot_max_z = max(eq.zbdry) * zoom

    ax.set_xlim(0, plot_max_x)
    ax.set_ylim(-plot_max_z, plot_max_z)

    show()


@cli.command("plot-psi", no_args_is_help=True)
@click.argument("filepath", type=click.Path(exists=True))
def plot_psi(filepath):
    """
    Plot the eqdsk psi map.

    matplotlib is required for plotting.
    """
    eq = EQDSKInterface.from_file(filepath, no_cocos=True)

    _fig, ax, show = _setup_plotting(eq)

    ax.set_title(
        "psi [V.s/rad (COCOS 1-8), V.s (COCOS 11-18)]",
        loc="right",
        fontdict={"fontsize": 10},
    )

    psi_x, psi_z = np.meshgrid(eq.x, eq.z, indexing="ij")

    if eq.psi.shape[0] != psi_x.shape[0]:
        eqdsk_warn(
            f"psi shape {eq.psi.shape} does match coords shape {psi_x.shape},"
            " attempting to swap..."
        )
        eq.psi = np.swapaxes(eq.psi, 0, 1)
    cont = ax.contour(psi_x, psi_z, eq.psi, levels=30, cmap="viridis")
    ax.clabel(cont, inline=True, fontsize=8, fmt="%d", colors="black")

    ax.legend(["Magnetic center"], loc="upper right")

    plot_max_x = eq.xmag + (eq.xdim / 2)
    plot_max_z = eq.zmag + (eq.zdim / 2)

    ax.set_xlim(0, plot_max_x)
    ax.set_ylim(-plot_max_z, plot_max_z)

    show()


class COCOSOptionsHelp(click.Command):
    """COCOS options help formatter"""

    def format_help(self, ctx, formatter):
        """Format the help with the Known COCOS options"""
        self.help = self.help.format(
            "'" + "', '".join(KnownCOCOS.__members__.keys()) + "'"
        )
        super().format_help(ctx, formatter)


@cli.command("convert", no_args_is_help=True, cls=COCOSOptionsHelp)
@click.argument("filepath", type=click.Path(exists=True))
@click.option(
    "-fmt",
    "--format",
    type=click.Choice(["json", "eqdsk"]),
    default="json",
    help="Format to save the eqdsk file in.",
)
@click.option(
    "-f",
    "--from",
    "from_",
    help="COCOS format to read the eqdsk as.",
)
@click.option(
    "-t",
    "--to",
    help="COCOS format to convert the eqdsk to.",
)
@click.option(
    "-q",
    "--qpsi-sign",
    type=click.Choice(["1", "-1"]),
    help="Sign of qpsi, "
    "required if the eqdsk has no qpsi data and you are converting between COCOS.",
)
def convert(
    filepath: str,
    format: str,  # noqa: A002
    from_: str | None,
    to: str | None,
    qpsi_sign: str | None,
):
    """
    Conversion utilities for the eqdsk file.

    To save the file in a different format, use the -fmt (--format) option.

    To convert between COCOS versions, use the -f (--from) and -t (--to)
    options (both must be provided).

    If -f and -t are not provided, the file will be read without COCOS.

    Valid values for --from and --to are:

      Integers: [1, 8] and [11, 18]

      Strings: {}

    The specified "from" COCOS value must be valid for the eqdsk file.

    Use the `eqdsk show` command to see the valid COCOS's for file.

    The saved file will have _out suffixed to the filename.
    """  # noqa: DOC501
    if from_ and to:
        # does validation of from and to values
        cc_fr = COCOS(from_)
        cc_to = COCOS(to)
        qsp = None if qpsi_sign is None else int(qpsi_sign) == 1
        eq = EQDSKInterface.from_file(
            filepath,
            from_cocos=cc_fr.index,
            to_cocos=cc_to.index,
            qpsi_positive=qsp,
        )
    elif (from_ and to is None) or (from_ is None and to):
        raise click.BadParameter("Both --from and --to must be provided")
    else:
        eq = EQDSKInterface.from_file(filepath, no_cocos=True)

    output_path = Path(filepath).with_stem(f"{Path(filepath).stem}_out").as_posix()
    eq.write(output_path, file_format=format)



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Enum models for values used in the COCOS specification."""

from __future__ import annotations

from enum import Enum
from typing import Any


class ZeroOne(Enum):
    """An enum representing the values 0 and 1 for the 2pi exponent of Bp."""

    ZERO = 0
    ONE = 1

    def __sub__(self, other: Any) -> ZeroOne:
        """
        Returns
        -------
        :
            The difference of the values as a new ZeroOne object

        Raises
        ------
        TypeError
            if not a ZeroOne object
        """
        if type(other) is ZeroOne:
            return ZeroOne(self.value - other.value)
        raise TypeError(
            f"Cannot subtract {type(other)} from {type(self)}.",
        )


class Sign(Enum):
    """An enum representing the positive or negative sign of
    a COCOS parameter.
    """

    POSITIVE = 1
    NEGATIVE = -1

    @classmethod
    def _missing_(cls, value) -> Sign:
        """

        Returns
        -------
        :
            The Sign object corresponding to the value.

        Raises
        ------
        ValueError
            If the value is not a known Sign.
        """
        if isinstance(value, bool):
            return Sign.POSITIVE if value else Sign.NEGATIVE
        raise ValueError(f"'{value}' not a known Sign") from None

    def __mul__(self, other: Any) -> Sign | int:
        """
        Returns
        -------
        :
            The product of the sign with the other value.

        Notes
        -----
        - If it is another Sign, return the product of the values.
        - If it is a number, return the product of the value and the number.
        """
        if type(other) is Sign:
            return Sign(self.value * other.value)
        return self.value * other



# SPDX-FileCopyrightText: 2023-present The Bluemira Developers <https://github.com/Fusion-Power-Plant-Framework/bluemira>  # noqa: INP001
#
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Generate the API reference pages and navigation."""

from pathlib import Path

import mkdocs_gen_files
from mkdocs_gen_files.nav import Nav

nav = Nav()

root = Path(__file__).parent.parent
src = root / "eqdsk"

# With the navigation.sections feature enabled, this isn't necessary.
# nav["Overview",] = "overview.md"  # noqa: ERA001

for path in sorted(src.rglob("*.py")):
    module_path = path.relative_to(root).with_suffix("")
    doc_path = path.relative_to(root).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    is_init = False

    if parts[-1] == "__init__":
        is_init = True
        parts = parts[:-1]
    elif parts[-1] == "__main__" or parts[-1] == "_version" or parts[-1] == "cli":
        continue

    p = ".".join(parts)
    parts_p = (*parts, p) if is_init else (*parts[:-1], p)
    nav[parts_p] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        fd.write(f"::: {p}")

    mkdocs_gen_files.set_edit_path(full_doc_path, path.relative_to(root))

with mkdocs_gen_files.open("reference/overview.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())
